// automatic selection of pid parameters 

function autoSelectTemplate(zoneID)
{
    // parameter validation - ignore pseudo zone 0
    if ((zoneID == null) || (zoneID == '0')) {
        return;
    }

    // if heating is globally disabled do not run autoselection of pid controller
    var heatingEnabled = Property.getProperty('generalSettings/heatingEnabled');
    if (heatingEnabled == null || heatingEnabled == false) {

        // TODO:
        // what is the resulting state and config?
        // need to modify current controller settings?

        return;
    }

    // an un-configured zone is per default in mode 'autoselect'
    var pZone = Property.getNode('controllerConfig/' + zoneID);
    if (pZone == null) {
        storeParameter('controllerConfig/' + zoneID + '/zoneID', zoneID);
        storeParameter('controllerConfig/' + zoneID + '/definitionID', -1);
		storeParameter('controllerConfig/' + zoneID + '/autoselect', 1);
        Property.store();
    }

    // autoselect active for this zone?
    var autoselect = Property.getProperty('controllerConfig/' + zoneID + '/autoselect');
    if ((autoselect == null) || (autoselect <= 0)) {
        lInfo.logln('PID autoselection not enabled for zone ' + zoneID + ': ' + autoselect);
        return;
    }

    // does this zone have a temperature sensor? without sensor no pid control possible...
    var zone = getZoneByID(parseInt(zoneID));
    var sensorDevice = zone.getAssignedSensor(9);
    if (sensorDevice == null || !sensorDevice.dsuid) {
        lInfo.logln('PID autoselection in zone ' + zoneID + ' skipped: no temperature sensor assigned');

        // TODO:
        // what is the resulting state and config?
        // need to modify current controller settings?

        return;
    }

    // definitionID is the numerical index of the controller parameter template
    // maybe negative, -1 => nothing selected, -2 => inconstistent valve types in zone, -3 => all valves Unknown
    var definitionID = Property.getProperty('controllerConfig/' + zoneID + '/definitionID');
    var zoneDefinitionID = DefinitionIDs.AutoNothingSelected;
    if (definitionID != null && definitionID >= 0) {
        zoneDefinitionID = definitionID;
    }

    var zoneValveType = "";
    var allValvesUnknown = true;
    var selectionStatus = 0;
    var numValveDevices = 0;
    var numHeatingSensors = 0;

    lInfo.logln('PID autoselection algorithm running for zone ' + zoneID);
    var devices = zone.getDevices();
    devices.perform(function (device) {

        // inactive devices must be ignored
        if (device.isPresent <= 0) {
            return;
        }

        // device valve type property
        var valveProperty = device.getPropertyNode().getChild('ValveType');
        if (valveProperty == null) {
            return;
        }
        var valveType = valveProperty.getValue();
        lInfo.logln('PID autoselection device: ' + device.dsuid + ' has valve type: ' + valveType);

        // run template selection algorithm
        if (valveType != null && valveType != "") {
            numValveDevices ++;
        }
        if (valveType != null && valveType !== "Unknown") {
            allValvesUnknown = false;
            if (valveType == "Floor" || valveType == "Wall") {
                valveType = "FloorWall";
            }
            if (selectionStatus == 0) {
                // nothing else found so far, the first device setting is saved
                zoneValveType = valveType;
                selectionStatus = 1;
            }
            else if (selectionStatus == 1) {
                // inconsistent/different types?
                if (zoneValveType != valveType) {
                    zoneValveType = "ERROR";
                    selectionStatus = -1;
                }
            }
        }
    });

    var result = '';
    result += '/ valve result: ' + zoneValveType + ' /';
    result += '/ selection status: ' +  selectionStatus + ' /';
    if (allValvesUnknown) {
        result += '/ all valves unknown /';
    }
    lInfo.logln('PID autoselection result for zone ' + zoneID + ': ' + result);

    // if we found consistent valve types try to find a best-match in the configuration templates,
    // the controllerDefinitions/ have been loaded from the cloud on startup
    if (selectionStatus > 0) {
        var pControllerTemplates = Property.getNode('controllerDefinitions').getChildren();
        var n;
        var newDefinitionID = -1;
        for (n = 0; n < pControllerTemplates.length; n ++) {
            var templateType = pControllerTemplates[n].getChild('type').getValue();
            if (zoneValveType == 'Radiator' && templateType == 1) {
                newDefinitionID = pControllerTemplates[n].getChild('id').getValue();
                break;
            }
            if (zoneValveType == 'FloorWall' && templateType == 2) {
                newDefinitionID = pControllerTemplates[n].getChild('id').getValue();
                break;
            }
        }
        lInfo.logln('PID autoselection template id for zone ' + zoneID + ': old id ' + zoneDefinitionID + ' => ' + newDefinitionID);
        if (newDefinitionID < 0) {
            lWarn.logln('PID autoselection template not found for zone ' + zoneID + ' and heating type ' + zoneValveType);
        }
        else if (newDefinitionID == zoneDefinitionID) {
            lInfo.logln('PID autoselection => nothing to change');
        }
        else {
            new Event('heating-controller.controller-config', {
                actions : 'setControllerModePID',
                zoneID : '' + zoneID,
                definitionID: '' + newDefinitionID,
                autoselect: 1
            }).raise();
        }
    }

    // if not consistent disable a possibly running controller and save the status AutoSelectionFailed (used by the UI)
    if (selectionStatus == -1) {
        lInfo.logln('PID autoselection failure for zone ' + zoneID);
        var pDefinitionID = Property.getNode('controllerConfig/' + zoneID + '/definitionID');
        pDefinitionID.setValue(DefinitionIDs.AutoSelectionFailed);
        Property.store();

        new Event('heating-controller.controller-config', {
            actions : 'resetControllerMode',
            zoneID : '' + zoneID,
            autoselect: 1,
            definitionID: DefinitionIDs.AutoSelectionFailed,
        }).raise();
    }

    // if no selection possible save the status AutoNothingSelected or AutoSelectionPossible accordingly (used by the UI)
    if ((selectionStatus == 0) || allValvesUnknown) {
        var statusValue = DefinitionIDs.AutoNothingSelected;
        if ((numValveDevices > 0) && allValvesUnknown) {
            statusValue = DefinitionIDs.AutoSelectionPossible;
        }

        var pDefinitionID = Property.getNode('controllerConfig/' + zoneID + '/definitionID');
        pDefinitionID.setValue(statusValue);
        Property.store();

        new Event('heating-controller.controller-config', {
            actions : 'resetControllerMode',
            zoneID : '' + zoneID,
            autoselect: 1,
            definitionID: statusValue,
        }).raise();
    }
}

if ((raisedEvent.name == 'heating-controller.autoselect') || (raisedEvent.name == 'DeviceHeatingTypeChanged')) {
	try {
        autoSelectTemplate(raisedEvent.parameter.zoneID);
    }
    catch (e) {
        lWarn.logln('PID autoselection: error on event ' + raisedEvent.name + ': ' + e);
    }
}

else if (raisedEvent.name == 'DeviceEvent' && raisedEvent.parameter.action == 'moved') {
	try {
        autoSelectTemplate(raisedEvent.parameter.oldZoneID);
        autoSelectTemplate(raisedEvent.parameter.newZoneID);
    }
    catch (e) {
        lWarn.logln('PID autoselection: error on event ' + raisedEvent.name + ': ' + e);
    }
}

/*
else if (raisedEvent.name == 'DeviceEvent' && raisedEvent.parameter.action == 'deleted') {
	Property.load();
	try {
        autoSelectTemplate(oldZoneID);
    }
    catch (e) {
        lWarn.logln('Error on heating-controller.autoselect DeviceEvent deleted: ' + e);
    }
}
*/
